/*
  fanctl_tiny.ino, for the Fan-Tach-shtick project.
  
  Fan clone control, ATtiny version, 8MHz or faster (for serial port.)
  Use optosensor to read GPU fan RPM, then adjusts case fan speed.
  
  Uses EEPROM so set EESAVE=0 fuse to not erase EEPROM during programming (EESAVEenable.bat utility) or:
  avrdude avrdude.conf -v -pattiny85 -cusbtiny -U hfuse:w:0xD7:m
  
 
  Fan Hum:
  Using default PWM settings (490Hz?) makes the fan hum audibly.
  So move the PWM F < 100Hz, 30Hz or so works ok.

  User settings -- see service_menu()
  
Pins used:
  Function     PhysPin  PB  desc
  LED           5       PB0 opto sensor indicator
  Sensor in     6       PB1 AIN1, w/pullup
  Serial out    7       PB2 Tx
  Serial in     2       PB3 Rx
  PWM out       3       PB4 OC1B

*/

#include <avr/power.h>  //for optional 16MHz
#include <limits.h>
#include <EEPROM.h>     
#include "analogComp.h"
#include <SoftwareSerial.h>
const int rx = PB3;
const int tx = PB2;
//SoftwareSerial mySerial(rx, tx, true); // setup pins, polarity = true for direct
SoftwareSerial mySerial(rx,tx,false);   // setup pins, polarity = false for RS232 adapter

#define SERVICE_MENU_KEY '~'              //for service menu

//global variables
const byte LED = PB0;                   //debug LED
volatile uint16_t ir_cntr;              //IR opto pulse counter
uint8_t fan_pwm;                        //current fan PWM
unsigned long report_ms, ctl_update_ms; //timers
uint16_t rpm_m = 60;                    //rev/sec to rpm multiplier, typically 60. If sensor detects 2 triggers/rev, this is 30.

#define AVE_WINDOW  8
uint16_t  ave_cntr_a[AVE_WINDOW];       //simple averager


//control params item index
enum {
  T0,F0,
  T1,F1,
  H0,
  U0,
  CHK,
  LAST_CP
};
uint16_t control_params[LAST_CP];  //simple array holds params, written to EEPROM.

//**********************************************************
//interrupt raised by the analog comparator, a simple counter
void changeStatus() {
  ir_cntr++;
}

//**********************************************************
void setup() {
    if (F_CPU == 16000000) clock_prescale_set(clock_div_1); //in case using 16MHz clk
    pinMode(LED, OUTPUT); //LED pin as output
    pinMode(tx,OUTPUT);
    pinMode(rx,INPUT);
    
    analogComparator.setOn(INTERNAL_REFERENCE, AIN1); //AIN0 is internal ref, not connected externally. AIN1 connected to opto.
    analogComparator.enableInterrupt(changeStatus, FALLING); //either one works...
    PORTB |= 1<<PORTB1; //PB1 Pullup for opto power
    
    mySerial.begin(9600);
    delay(100);
    mySerial.print( F("Tiny Opto tach '") );
    mySerial.print(SERVICE_MENU_KEY);
    mySerial.println( F("' for entry") );
}

//**********************************************************
//calcs difference between 2 timestamps -- just uint w/rollover.
static unsigned long calc_ts_et(unsigned long ts, unsigned long last_ts)
{
  if ( ts >= last_ts )
  {
    return ( ts - last_ts );
  }
  else
  {
    return ( 1 + ts - (ULONG_MAX - last_ts) );
  }
}

//**********************************************************
//reads delimited (or \n) bytes, no timeout, some useful small limit...
String myReadSerialDelim( char delimiter ) {
  String str = "";
  int c=0;
  //wait for input
  while (mySerial.available() <= 0)
      ;
  do {
    while (mySerial.available () > 0) {
      c = mySerial.read ();
      str += char(c);
    }
  } while ( c != delimiter && c != '\n' && str.length() < 24 );
  return str;
}


//**********************************************************
// configure 8-bit PWM on PB4, under 60Hz or so. 8,000,000/256/Pre = PWM clk
void pwm_direct8( byte pwm )
{
  pinMode(PB4, OUTPUT);  //pwm out
  if(pwm == 0)
  {
    digitalWrite(PB4, 0); //force off
  }
  
  // Configure counter/timer1 PWM1B for fast PWM on PB4, OCR1B
  GTCCR = 1<<PWM1B | 2<<COM1B0;
  //set Prescale for  <60Hz
  TCCR1 = 3<<COM1A0 | (1<<CS13) | (0<<CS12) | (1<<CS11) | (1<<CS10);  //30Hz
  OCR1B = pwm;  
}

//for testing, show raw counter
void comp_report(void)
{
  while( !keypress_test() )
  {
    mySerial.println( ir_cntr );
    delay(100);
  }
}

void report_control_params(void)
{
  uint8_t chk=0;
  
  mySerial.println( F("control_params:"));
  EEPROM.get(0, control_params);
  for(int i=0; i<CHK; i++)
  {
    chk ^= control_params[i];
    mySerial.println( control_params[i] );
  }
  chk++;
  mySerial.print( control_params[CHK] );
  mySerial.print( F(" CHK "));
  mySerial.println( control_params[CHK] == chk ? F("Pass"): F("Fail") );
}

//**********************************************************
//returns 1 if SERVICE_MENU_KEY entered
int keypress_test(void)
{
  if ( mySerial.available() > 0 )
  {
    if (  mySerial.read() == SERVICE_MENU_KEY )
    {
      delay(100);
      //eat NL
      while ( mySerial.available() > 0 )
      {
        mySerial.read();
      }
      return (1);
    }
  }
  return (0);
}

enum
{
  MENU_START,
  MENU_A,
  MENU_B,
  MENU_C,
  MENU_D,
  MENU_E,
  MENU_F,
  MENU_G,
};

void service_menu( void )
{
  while (1)
  {
    /* Display Menu on Terminal Window */
    mySerial.println( F("\n==== Fan Clone ====") ) ;
    mySerial.print(MENU_A);   mySerial.println( F(": Set PWM") );
    mySerial.print(MENU_B);   mySerial.println( F(": Show IR Cntr") );
    mySerial.print(MENU_C);   mySerial.println( F(": Set RPM Multiply") );
    mySerial.print(MENU_D);   mySerial.println( F(": Set Control Parameters") );
    mySerial.print(MENU_E);   mySerial.println( F(": Show Control Parameters") );

    mySerial.println( F("Enter selection [q]: ") );

    while (mySerial.available() <= 0)           //timeout enabled once a keypress
      ;
    String choice = mySerial.readStringUntil('\n');

    if ( (choice.charAt(0) | 0x20) == 'q' )   //tolower
    {
      mySerial.println( F("Bye!") );
      return;
    }

    switch ( choice.toInt() )
    {
      case MENU_A:  //manually set fan PWM
      {
        mySerial.println( F("Enter PWM:"));
        fan_pwm = myReadSerialDelim('\n').toInt();
        pwm_direct8( fan_pwm );
        mySerial.println(fan_pwm);
      }
      break;
      
      case MENU_B:
        comp_report();  //show cntr
      break;
      
      case MENU_C:
        mySerial.println( F("Enter Multiplier:"));    // in case sensor triggers other than 1 per rev. For debugging. 
        rpm_m = myReadSerialDelim('\n').toInt();
        break;
        
      case MENU_D:
      {
        
        // Control items.
        // 1) Fan On, Target min RPM, fan min PWM (0..255), [T0:F0]   
        // 2) Fan Max, Target max RPM, fan max PWM,         [T1:F1]   
        // 3) hysterisis RPM, say 400                       [H0]
        // 4) update rate, seconds, 5 works ok.             [U0]         
        mySerial.println( F("Enter Parameters T0:F0:T1:F1:H0:U0"));
        uint8_t chk=0;
        for(int i=0; i<CHK; i++)
        {
          control_params[i] = myReadSerialDelim(':').toInt();
          chk ^= control_params[i];
        }
        chk++;
        control_params[CHK] = chk;    //simple xor+1 checksum

        mySerial.println( F("Saved"));
        EEPROM.put(0, control_params);
        report_control_params();
      }
      break;

      case MENU_E:
        report_control_params();
      break;

    }
  }
}


/********* main loop  *****************/
void loop() {
  uint16_t  ave_cnt = 0;
  uint16_t  old_cnt = 0;
  uint8_t   ave_idx = 0;
  uint8_t   h_flag = 0;
  
  
  ctl_update_ms = report_ms = millis();
  fan_pwm = 0;
  pwm_direct8( fan_pwm );

  report_control_params();
  
  while(1)
  {
    digitalWrite(LED, ir_cntr & 1);    //show sensor triggers
    
    if ( keypress_test() )
    {
      service_menu();

      //reset reporting
      report_ms = millis();
      ir_cntr=0;
    }


    //update Fan speed with hysterisis
    if ( calc_ts_et( millis(), ctl_update_ms) >= 1000L* control_params[U0] )
    {
      ctl_update_ms = millis();

      if( ave_cnt >= control_params[T0]+control_params[H0] )
      {
        //allow lower hysterisis
        h_flag = 1;
      } 
      else if( ave_cnt < control_params[T0] )
      {
        //turn off lower hysterisis
        h_flag = 0;
      }

      //adjust fan speed
      if( ave_cnt >= control_params[T0] )
      {   
        int16_t delta_cnt = (int16_t)ave_cnt - (int16_t)old_cnt;      //change since last update
        
        if( ( h_flag == 0 )                                   ||    //no lower hysterisis, update always
            ( delta_cnt > 0 )                                 ||    //allow increase always
            ( h_flag==1 && delta_cnt < -control_params[H0] )    )   //allow decrease after hysterisis
        {
          //map w/contraint limits
          int v = constrain(ave_cnt, control_params[T0], control_params[T1]);
          fan_pwm = map( v, control_params[T0], control_params[T1], control_params[F0], control_params[F1]);
          old_cnt = ave_cnt;
        }
      }
      else
      {
        //turn off
        fan_pwm = 0;
        h_flag = 0;
      }
      pwm_direct8( fan_pwm );
      mySerial.print( F("Fan PWM: ") ); 
      mySerial.println( fan_pwm ); 
    }

    

    //simple linear Fan ctl (optional)
    if (0 && calc_ts_et( millis(), ctl_update_ms) >= 1000L* control_params[U0] )
    {
      ctl_update_ms = millis();
      
      //if target min RPM reached turn on fan, linear map
      if( ave_cnt >= control_params[T0] )
      {
        int v = constrain(ave_cnt, control_params[T0], control_params[T1]);
        fan_pwm = map( v, control_params[T0], control_params[T1], control_params[F0], control_params[F1]);
      }
      else 
      {
        fan_pwm = 0;  //otherwise force off
      }
      
      pwm_direct8( fan_pwm );
      mySerial.print( F("Fan PWM: ") ); 
      mySerial.println( fan_pwm ); 
    }

    //report opto rpm, update average
    if ( calc_ts_et( millis(), report_ms) >= 1000 )
    {
      uint16_t rpm;    
      
      rpm = ir_cntr * rpm_m;    //simple RPM calc based on report freq, sensor blips/sec.

      //simple averager, note upper limit of 8k rpm or 16-bit rollover
      ave_cntr_a[ave_idx++] = rpm;         
      if( ave_idx >= AVE_WINDOW ) ave_idx = 0;
      
      uint16_t sum=0;
      for( int i=0; i<AVE_WINDOW; i++)
      {
        sum += ave_cntr_a[i];
      }
      ave_cnt = sum/AVE_WINDOW;
      
      mySerial.print( F("Target RPM:") ); 
      mySerial.print( rpm ); 
      mySerial.print( F("  Ave: ") ); 
      mySerial.println( ave_cnt );
      
      //reset for next report
      report_ms = millis();
      ir_cntr=0;
    }
  }
}
